@@ -1,4 +1,5 @@
-import cPickle
+from beaker._compat import pickle
+
 import logging
 import pickle
 from datetime import datetime
@@ -79,7 +80,7 @@ class SqlaNamespaceManager(OpenResourceNamespaceManager):
             self._is_new = False
             try:
                 self.hash = result['data']
-            except (IOError, OSError, EOFError, cPickle.PickleError,
+            except (IOError, OSError, EOFError, pickle.PickleError,
                     pickle.PickleError):
                 log.debug("Couln't load pickle data, creating new storage")
                 self.hash = {}
