@@ -9,6 +9,12 @@ from beaker.cache import clsmap
 from beaker.exceptions import BeakerException, InvalidCryptoBackendError
 from beaker.cookie import SimpleCookie
 
+
+months = (None, "Jan", "Feb", "Mar", "Apr", "May", "Jun",
+          "Jul", "Aug", "Sep", "Oct", "Nov", "Dec")
+weekdays = ("Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun")
+
+
 __all__ = ['SignedCookie', 'Session', 'InvalidSignature']
 
 
@@ -86,7 +92,27 @@ class SignedCookie(SimpleCookie):
         return str(val), ("%s%s" % (sig, val))
 
 
-class Session(dict):
+class _ConfigurableSession(dict):
+    """Provides support for configurable Session objects.
+    
+    Provides a way to ensure some properties of sessions
+    are always available with pre-configured values
+    when they are not available in the session cookie itself.
+    """
+
+    def __init__(self, cookie_domain=None, cookie_path='/'):
+        self._config = {
+            '_domain': cookie_domain,
+            '_path': cookie_path
+        }
+
+    def clear(self):
+        """Clears Session data. Preserves session configuration."""
+        super(_ConfigurableSession, self).clear()
+        self.update(self._config)
+
+
+class Session(_ConfigurableSession):
     """Session object that uses container package for storage.
 
     :param invalidate_corrupt: How to handle corrupt data when loading. When
@@ -127,6 +153,8 @@ class Session(dict):
                                to keep backward compatibility with sessions generated before 1.8.0
                                set this to 48.
     :param crypto_type: encryption module to use
+    :param samesite: SameSite value for the cookie -- should be either 'Lax',
+                     'Strict', or None.
     """
     def __init__(self, request, id=None, invalidate_corrupt=False,
                  use_cookies=True, type=None, data_dir=None,
@@ -135,8 +163,15 @@ class Session(dict):
                  data_serializer='pickle', secret=None,
                  secure=False, namespace_class=None, httponly=False,
                  encrypt_key=None, validate_key=None, encrypt_nonce_bits=DEFAULT_NONCE_BITS,
-                 crypto_type='default',
+                 crypto_type='default', samesite='Lax',
                  **namespace_args):
+        _ConfigurableSession.__init__(
+            self,
+            cookie_domain=cookie_domain,
+            cookie_path=cookie_path
+        )
+        self.clear()
+        
         if not type:
             if data_dir:
                 self.type = 'file'
@@ -156,6 +191,15 @@ class Session(dict):
         if timeout and not save_accessed_time:
             raise BeakerException("timeout requires save_accessed_time")
         self.timeout = timeout
+
+        # If a timeout was provided, forward it to the backend too, so the backend
+        # can automatically expire entries if it's supported.
+        if self.timeout is not None:
+            # The backend expiration should always be a bit longer than the
+            # session expiration itself to prevent the case where the backend data expires while
+            # the session is being read (PR#153). 2 Minutes seems a reasonable time.
+            self.namespace_args['timeout'] = self.timeout + 60 * 2
+
         self.save_atime = save_accessed_time
         self.use_cookies = use_cookies
         self.cookie_expires = cookie_expires
@@ -163,12 +207,11 @@ class Session(dict):
         self._set_serializer(data_serializer)
 
         # Default cookie domain/path
-        self._domain = cookie_domain
-        self._path = cookie_path
         self.was_invalidated = False
         self.secret = secret
         self.secure = secure
         self.httponly = httponly
+        self.samesite = samesite
         self.encrypt_key = encrypt_key
         self.validate_key = validate_key
         self.encrypt_nonce_size = get_nonce_size(encrypt_nonce_bits)
@@ -233,15 +276,24 @@ class Session(dict):
 
     def _set_cookie_values(self, expires=None):
         self.cookie[self.key] = self.id
-        if self._domain:
-            self.cookie[self.key]['domain'] = self._domain
+        if self.domain:
+            self.cookie[self.key]['domain'] = self.domain
         if self.secure:
             self.cookie[self.key]['secure'] = True
+        if self.samesite:
+            self.cookie[self.key]['samesite'] = self.samesite
         self._set_cookie_http_only()
-        self.cookie[self.key]['path'] = self._path
+        self.cookie[self.key]['path'] = self.path
 
         self._set_cookie_expires(expires)
 
+    @staticmethod
+    def serialize_cookie_date(v):
+        v = v.timetuple()
+        r = time.strftime("%%s, %d-%%s-%Y %H:%M:%S GMT", v)
+        return r % (weekdays[v[6]], months[v[1]])
+
+        
     def _set_cookie_expires(self, expires):
         if expires is None:
             expires = self.cookie_expires
@@ -261,12 +313,15 @@ class Session(dict):
             self.cookie[self.key]['expires'] = ''
             return True
         self.cookie[self.key]['expires'] = \
-            expires_date.strftime("%a, %d-%b-%Y %H:%M:%S GMT")
+            self.serialize_cookie_date(expires_date)
         return expires_date
 
     def _update_cookie_out(self, set_cookie=True):
         self._set_cookie_values()
-        self.request['cookie_out'] = self.cookie[self.key].output(header='')
+        cookie_out = self.cookie[self.key].output(header='')
+        if not isinstance(cookie_out, str):
+            cookie_out = cookie_out.encode('latin1')
+        self.request['cookie_out'] = cookie_out
         self.request['set_cookie'] = set_cookie
 
     def _set_cookie_http_only(self):
@@ -293,20 +348,20 @@ class Session(dict):
         return self['_creation_time']
 
     def _set_domain(self, domain):
-        self['_domain'] = self._domain = domain
+        self['_domain'] = domain
         self._update_cookie_out()
 
     def _get_domain(self):
-        return self._domain
+        return self['_domain']
 
     domain = property(_get_domain, _set_domain)
 
     def _set_path(self, path):
-        self['_path'] = self._path = path
+        self['_path'] = path
         self._update_cookie_out()
 
     def _get_path(self):
-        return self._path
+        return self.get('_path', '/')
 
     path = property(_get_path, _set_path)
 
@@ -420,9 +475,6 @@ class Session(dict):
                 # Update the current _accessed_time
                 session_data['_accessed_time'] = now
 
-                # Set the path if applicable
-                if '_path' in session_data:
-                    self._path = session_data['_path']
                 self.update(session_data)
                 self.accessed_dict = session_data.copy()
         finally:
@@ -547,18 +599,26 @@ class CookieSession(Session):
                                otherwise invalid data will cause an exception.
     :type invalidate_corrupt: bool
     :param crypto_type: The crypto module to use.
+    :param samesite: SameSite value for the cookie -- should be either 'Lax',
+                     'Strict', or None.
     """
     def __init__(self, request, key='beaker.session.id', timeout=None,
                  save_accessed_time=True, cookie_expires=True, cookie_domain=None,
                  cookie_path='/', encrypt_key=None, validate_key=None, secure=False,
                  httponly=False, data_serializer='pickle',
                  encrypt_nonce_bits=DEFAULT_NONCE_BITS, invalidate_corrupt=False,
-                 crypto_type='default',
+                 crypto_type='default', samesite='Lax',
                  **kwargs):
+        _ConfigurableSession.__init__(
+            self,
+            cookie_domain=cookie_domain,
+            cookie_path=cookie_path
+        )
+        self.clear()
 
         self.crypto_module = get_crypto_module(crypto_type)
 
-        if not self.crypto_module.has_aes and encrypt_key:
+        if encrypt_key and not self.crypto_module.has_aes:
             raise InvalidCryptoBackendError("No AES library is installed, can't generate "
                                             "encrypted cookie-only Session.")
 
@@ -573,8 +633,7 @@ class CookieSession(Session):
         self.request['set_cookie'] = False
         self.secure = secure
         self.httponly = httponly
-        self._domain = cookie_domain
-        self._path = cookie_path
+        self.samesite = samesite
         self.invalidate_corrupt = invalidate_corrupt
         self._set_serializer(data_serializer)
 
@@ -611,7 +670,6 @@ class CookieSession(Session):
                 if cookie_data is InvalidSignature:
                     raise BeakerException("Invalid signature")
                 self.update(self._decrypt_data(cookie_data))
-                self._path = self.get('_path', '/')
             except Exception as e:
                 if self.invalidate_corrupt:
                     util.warn(
@@ -641,18 +699,17 @@ class CookieSession(Session):
 
     def _set_domain(self, domain):
         self['_domain'] = domain
-        self._domain = domain
 
     def _get_domain(self):
-        return self._domain
+        return self['_domain']
 
     domain = property(_get_domain, _set_domain)
 
     def _set_path(self, path):
-        self['_path'] = self._path = path
+        self['_path'] = path
 
     def _get_path(self):
-        return self._path
+        return self['_path']
 
     path = property(_get_path, _set_path)
 
@@ -691,17 +748,20 @@ class CookieSession(Session):
         if expires is not None:
             self['_expires'] = expires
 
-        if '_domain' in self:
-            self.cookie[self.key]['domain'] = self['_domain']
-        elif self._domain:
-            self.cookie[self.key]['domain'] = self._domain
+        if self.domain:
+            self.cookie[self.key]['domain'] = self.domain
         if self.secure:
             self.cookie[self.key]['secure'] = True
+        if self.samesite:
+            self.cookie[self.key]['samesite'] = self.samesite
         self._set_cookie_http_only()
 
         self.cookie[self.key]['path'] = self.get('_path', '/')
 
-        self.request['cookie_out'] = self.cookie[self.key].output(header='')
+        cookie_out = self.cookie[self.key].output(header='')
+        if not isinstance(cookie_out, str):
+            cookie_out = cookie_out.encode('latin1')
+        self.request['cookie_out'] = cookie_out
         self.request['set_cookie'] = True
 
     def delete(self):
@@ -801,6 +861,7 @@ class SessionObject(object):
 
         Always saves the whole session if save() or delete() have been called.
         If they haven't:
+
         - If autosave is set to true, saves the the entire session regardless.
         - If save_accessed_time is set to true or unset, only saves the updated
           access time.
